library nui_error_handler;

import 'dart:async';
import 'dart:typed_data';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:nui_core/code/nui_device_util.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_database/models/enums/nui_query_order.dart';
import 'package:nui_database/nui_database.dart';
import 'package:nui_error_handler/nui_error_database.dart';
import 'package:nui_error_handler/nui_error_handler_action.dart';
import 'package:nui_error_handler/nui_error_models.dart';
import 'package:nui_error_handler/nui_error_ui_components.dart';
import 'package:nui_error_handler/nui_error_utils.dart';
import 'package:http/http.dart' as http;

export 'package:nui_error_handler/nui_error_database.dart';
export 'package:nui_error_handler/nui_error_ui_components.dart';
export 'package:nui_error_handler/nui_error_utils.dart';
export 'package:nui_error_handler/nui_error_handler_action.dart';
export 'package:nui_error_handler/nui_error_models.dart';

typedef OnExceptionListener = void Function(Object error, StackTrace stackTrace);
typedef ErrorHandlerGetter = NUIErrorHandlerBuilder Function();
typedef OnInitListener = void Function();

void runNUIAppWithErrorHandler(NUIApp app, ErrorHandlerGetter errorHandlerBuilder, {OnInitListener initListener, OnExceptionListener exceptionListener}) async{
  Future<NUIErrorHandler> errorBuilder;
  runZonedGuarded<Future<void>>(() async {
    runApp(app);
    if(initListener != null) initListener();
  }, (Object error, StackTrace stackTrace) async {
    if(exceptionListener != null) exceptionListener(error, stackTrace);
    if(errorBuilder == null){
      errorBuilder = errorHandlerBuilder != null ? errorHandlerBuilder().build() : null;
    }
    if(error.toString().toLowerCase().contains("betterplayer")){
      //Do nothing
    }
    else {
      final errorHandler = await errorBuilder;
      String screenCode = NUINavigator.getCurrentScreenCode();
      final saveError = await errorHandler.saveErrorByException(
          error, stackTrace, screenCode: screenCode);
      if (saveError) {
        if (errorHandler._showErrorOnEncounter == true) {
          errorHandler.showLastErrorLog(null);
        }
      }
    }
  });
  FlutterError.onError = (FlutterErrorDetails details){
    Zone.current.handleUncaughtError(details.exception, details.stack);
  };
}

enum NUIErrorCode{
  NO_INTERNET_CONNECTION,
  FAILED_TO_GET_PERMISSION,
  OUT_OF_STORAGE,
  OUT_OF_MEMORY,
  VERSION_OUTDATED,
  FAILED_TO_ACCESS_DATABASE,
}

extension NUIErrorCodeExt on NUIErrorCode{
  int value(){
    switch(this){
      case NUIErrorCode.NO_INTERNET_CONNECTION: return 1001;
      case NUIErrorCode.FAILED_TO_GET_PERMISSION: return 1002;
      case NUIErrorCode.OUT_OF_STORAGE: return 1003;
      case NUIErrorCode.OUT_OF_MEMORY: return 1004;
      case NUIErrorCode.VERSION_OUTDATED: return 1005;
      case NUIErrorCode.FAILED_TO_ACCESS_DATABASE: return 1006;
    }
  }
}

enum NUIErrorActionType{
  LOG,
  TOAST,
  POPUP,
  REDIRECT
}

NUIErrorActionType errorActionValueOf(String value){
  switch(value){
    case NUIErrorHandler.ACTION_LOG : return NUIErrorActionType.LOG;
    case NUIErrorHandler.ACTION_TOAST : return NUIErrorActionType.TOAST;
    case NUIErrorHandler.ACTION_POPUP : return NUIErrorActionType.POPUP;
    case NUIErrorHandler.ACTION_REDIRECT : return NUIErrorActionType.REDIRECT;
    default: return null;
  }
}

extension NUIErrorActionTypeExt on NUIErrorActionType{
  String value(){
    switch(this){
      case NUIErrorActionType.LOG : return NUIErrorHandler.ACTION_LOG;
      case NUIErrorActionType.TOAST : return NUIErrorHandler.ACTION_TOAST;
      case NUIErrorActionType.POPUP : return NUIErrorHandler.ACTION_POPUP;
      case NUIErrorActionType.REDIRECT : return NUIErrorHandler.ACTION_REDIRECT;
      default: return null;
    }
  }
}

enum NUIErrorType{
  NON_SPECIFIED,
  CRASH, //App crashes
  HTTP, //Errors received on HTTP calls
  LOGIC, //Exceptions during business logic
  UI, //Exceptions due to UI handling
  DATABASE, //Exceptions from database
  HARDWARE, //Exceptions while accessing hardware (eg. Camera, Storage)
}

NUIErrorType errorTypeValueOf(String value){
  switch(value) {
    case "Non-Specified" : return NUIErrorType.NON_SPECIFIED;
    case "Crash" : return NUIErrorType.CRASH;
    case "Database" : return NUIErrorType.DATABASE;
    case "HTTP" : return NUIErrorType.HTTP;
    case "Logic" : return NUIErrorType.LOGIC;
    case "UI" : return NUIErrorType.UI;
    case "Hardware" : return NUIErrorType.HARDWARE;
    default: return null;
  }
}

extension NUIErrorTypeExt on NUIErrorType{
  String value() {
    switch (this) {
      case NUIErrorType.NON_SPECIFIED:
        return "Non-Specified";
      case NUIErrorType.CRASH:
        return "Crash";
      case NUIErrorType.DATABASE:
        return "Database";
      case NUIErrorType.HTTP:
        return "HTTP";
      case NUIErrorType.LOGIC:
        return "Logic";
      case NUIErrorType.UI:
        return "UI";
      case NUIErrorType.HARDWARE:
        return "Hardware";
      default:
        return null;
    }
  }
}

abstract class NUIErrorHandlerInterface{
  Future<bool> saveErrorLog(NUIErrorLog log);
  Future<bool> saveErrorByException(Object error, StackTrace stackTrace, {String screenCode});
  Future<void> showLastErrorLog(BuildContext context);
  Future<void> showErrorLogDetail(BuildContext context, NUIErrorLog log);
  Future<void> showErrorLogs(BuildContext context);
  Future<List<NUIErrorLog>> getErrorLogs();
  Future<NUIErrorLog> getLastErrorLog();
  Future<void> processError(BuildContext context, String code, {String module, NUIErrorType errorType});
  Future<void> _fetchErrorMapping();
  Future<void> _fetchErrorMappingLocal();
}

class NUIErrorHandler implements NUIErrorHandlerInterface{
  static const String ACTION_LOG = "Log";
  static const String ACTION_TOAST = "Toast";
  static const String ACTION_POPUP = "Pop-Up";
  static const String ACTION_REDIRECT = "Redirect";

  NUIErrorHandler._builder(NUIErrorHandlerBuilder builder) {
    _mappingUrl = builder._mappingUrl;
    _appVersion = builder._appVersion;
    _module = builder._module;
    _redirect = builder._redirect;
    _popUp = builder._popUpCallback;
    _showErrorOnEncounter = builder._showErrorDetailOnCrash;
    _filterByModule = builder._filterByModule;
    _getDeviceInfo();
    _fetchErrorMapping();
    NUIErrorHandler._errorHandlers[_module ?? _DEFAULT_MODULE] = this;
  }

  static const String MODULE = "NUIErrorHandler";
  static const String _DEFAULT_MODULE = "nui_default_error_module";
  static Map<String, NUIErrorHandler> _errorHandlers = Map();

  String _mappingUrl;
  String _appVersion;
  String _module;
  bool _filterByModule;
  NUIDeviceInfo _deviceInfo;
  bool _showErrorOnEncounter;
  ErrorRedirectCallback _redirect;
  ErrorPopUpCallback _popUp;

  static NUIErrorHandler get({String module}){
    final targetModule = module ?? _DEFAULT_MODULE;
    final errorHandler = _errorHandlers[targetModule];

    if(errorHandler == null){
      throwNUI(MODULE, "Instance has not been initialized for module: $targetModule");
    }
    return errorHandler;
  }

  Future<NUIDeviceInfo> _getDeviceInfo() async{
    if(_deviceInfo == null) {
      _deviceInfo = await NUIDeviceUtil.getDeviceInfo();
    }
    return _deviceInfo;
  }

  @override
  Future<void> _fetchErrorMappingLocal() async{
    List<NUIErrorMapping> errorMappings;
    try {
      ByteData data = await rootBundle.load("assets/error_mapping.xlsx");
      final bytes = data.buffer.asUint8List(
          data.offsetInBytes, data.lengthInBytes);
      if (bytes != null) {
        errorMappings = NUIErrorUtils.readErrorMappingFromExcel(bytes);
        await NUIErrorHandlerDatabase.get().clear(NUIErrorMappingEntity());
        await NUIErrorHandlerDatabase.get().saveList(errorMappings);
      }
    }catch(e){}

    if(errorMappings.isNullOrEmpty()){
      try{
        String data = await rootBundle.loadString("assets/error_mapping.json");
        errorMappings = NUIErrorUtils.readErrorMappingFromJson(value: data);
      }catch(e){}
    }

    if(!errorMappings.isNullOrEmpty()) {
      await NUIErrorHandlerDatabase.get().clear(NUIErrorMappingEntity());
      await NUIErrorHandlerDatabase.get().saveList(errorMappings);
    }
  }

  Future<Uint8List> _downloadFromURL(String url) async{
    try {
      http.Client client = new http.Client();
      var req = await client.get(Uri.parse(url));
      var bytes = req.bodyBytes;
      return bytes;
    }catch(e){
      return null;
    }
  }

  @override
  Future<void> _fetchErrorMapping() async {
    if (!isNullOrEmpty(_mappingUrl)){
      logNUI(MODULE, "Fetching error mapping from url : $_mappingUrl");
      final bytes = await _downloadFromURL(_mappingUrl);
      if (bytes != null) {
        final errorMappings = NUIErrorUtils.readErrorMappingFromExcel(bytes);
        await NUIErrorHandlerDatabase.get().clear(NUIErrorMappingEntity());
        await NUIErrorHandlerDatabase.get().saveList(errorMappings);
        return;
      }
    }

    logNUI(MODULE, "Fetching error mapping from local");
    await _fetchErrorMappingLocal();
  }

  @override
  Future<List<NUIErrorLog>> getErrorLogs() async{
    final errorLog = await NUIErrorHandlerDatabase.get().getList(NUIErrorLogEntity(), orderBy: "dateTime", order: NUIQueryOrder.DESC, limit: 50);
    return errorLog;
  }

  @override
  Future<NUIErrorLog> getLastErrorLog() async{
    final errorLog = await NUIErrorHandlerDatabase.get().getOne(NUIErrorLogEntity(), orderBy: "dateTime", order: NUIQueryOrder.DESC);
    return errorLog;
  }

  @override
  Future<void> processError(BuildContext context, String code, {String module, NUIErrorType errorType}) async{
    final errorMappingByCode = await NUIErrorHandlerDatabase.get().getOne(NUIErrorMappingEntity(), condition: (query){
      String filterModule = module ?? (_filterByModule == true ? _module : null);
      if(filterModule.isNullOrEmpty()){
        return query.where(field: "code", type: NUIEntType.STRING, condition: NUIDBQueryType.EQUALS, value: code);
      }
      return query.where(field: "code", type: NUIEntType.STRING, condition: NUIDBQueryType.EQUALS, value: code).where(field: "module", type: NUIEntType.STRING, condition: NUIDBQueryType.EQUALS, value: filterModule);
    });
    if(errorMappingByCode == null){
      logNUI(MODULE, "No error mapping found for code : $code and module : $module, no action will be taken by the error handler");
      return;
    }

    logNUI(MODULE, "Found error mapping found for code : $code, type: ${errorMappingByCode.actionType}");
    switch(errorMappingByCode.actionType){
      case NUIErrorHandler.ACTION_LOG : {

        break;
      }
      case NUIErrorHandler.ACTION_TOAST : {
        toast(errorMappingByCode.message);
        break;
      }
      case NUIErrorHandler.ACTION_POPUP : {
        if(_popUp != null){
          final dialog = _popUp(errorMappingByCode);
          triggerNUIDialog(dialog);
        }
        else{
          logNUI(NUIErrorHandler.MODULE, "Sending pop-up action to screen");
          triggerNUIDialog(NUIDialogItem(content: (context) => NUIErrorPopUpDialog(errorMappingByCode, listener: _redirect), opacity: 0.4));
        }
        break;
      }
      case NUIErrorHandler.ACTION_REDIRECT : {
        if(_redirect != null) _redirect(context, errorMappingByCode.actionCode, errorMappingByCode);
        break;
      }
    }
  }

  @override
  Future<bool> saveErrorByException(Object error, StackTrace stackTrace, {String screenCode}) async{
    try {
      String title = error.toString();
      String description = stackTrace.toString();
      NUIDeviceInfo deviceInfo = await _getDeviceInfo();
      NUIErrorLog log = NUIErrorLog(
          id: randomUUID(),
          type: NUIErrorType.CRASH.value(),
          title: title,
          description: description,
          appVersion: _appVersion,
          dateTime: DateTime.now(),
          module: _module,
          code: deviceInfo.version,
          platform: deviceInfo.platform,
          model: deviceInfo.model,
          screenCode: screenCode
      );
      return saveErrorLog(log);
    }catch(e){
      logNUI(MODULE, "Failed to save error log with error : $e");
      return false;
    }
  }

  @override
  Future<bool> saveErrorLog(NUIErrorLog log) async {
    NUIDeviceInfo deviceInfo = await _getDeviceInfo();
    if(isNullOrEmpty(log.id)) log.id = randomUUID();
    log.appVersion = _appVersion;
    if(isNullOrEmpty(log.dateTime)) log.dateTime = DateTime.now();
    log.module = _module;
    log.code = deviceInfo.version;
    log.platform = deviceInfo.platform;
    log.model = deviceInfo.model;
    return NUIErrorHandlerDatabase.get().save(log, recordDate: true);
  }

  @override
  Future<void> showErrorLogs(BuildContext context) async{
    final logs = await getErrorLogs();
    if(!isNullOrEmpty(logs)){
      showEntityTable(context, NUIErrorLogEntity(), logs, listener: (data){
        if(data is NUIErrorLog){
          showErrorLogDetail(context, data);
        }
      });
    }
  }

  @override
  Future<void> showErrorLogDetail(BuildContext context, NUIErrorLog log) async {
    if(NUIErrorDetailDialog.showing == true) return;

    if(context != null) {
      final dialog = showNUIDialog(
          context: context,
          content: (context) => NUIErrorDetailDialog(log),
          dismissible: true,
          useSafeArea: false,
          alignment: Alignment.center
      );
    }
    else{
      triggerNUIDialog(NUIDialogItem(content: (context) => NUIErrorDetailDialog(log), dismissible: true, useSafeArea: false, alignment: Alignment.center));
    }
  }

  @override
  Future<void> showLastErrorLog(BuildContext context) async {
    final errorLog = await getLastErrorLog();
    if(errorLog == null){
      logNUI(MODULE, "No error logs found");
      return;
    }
    showErrorLogDetail(context, errorLog);
  }
}

typedef ErrorRedirectCallback = void Function(BuildContext context, String actionCode, NUIErrorMapping errorMapping);
typedef ErrorPopUpCallback = NUIDialogItem Function(NUIErrorMapping errorMapping);

class NUIErrorHandlerBuilder{
  String _mappingUrl;
  String _appVersion;
  String _module;
  bool _filterByModule;
  ErrorRedirectCallback _redirect;
  bool _showErrorDetailOnCrash;
  ErrorPopUpCallback _popUpCallback;

  NUIErrorHandlerBuilder appVersion(String version){
    this._appVersion = version;
    return this;
  }

  NUIErrorHandlerBuilder filterByModule(bool filter){
    this._filterByModule = filter;
    return this;
  }

  NUIErrorHandlerBuilder showErrorOnEncounter(bool showOnEncounter){
    this._showErrorDetailOnCrash = showOnEncounter;
    return this;
  }

  NUIErrorHandlerBuilder redirect(ErrorRedirectCallback redirect){
    this._redirect = redirect;
    return this;
  }

  NUIErrorHandlerBuilder popUp(ErrorPopUpCallback popUp){
    this._popUpCallback = popUp;
    return this;
  }

  NUIErrorHandlerBuilder module(String module){
    this._module = module;
    return this;
  }

  NUIErrorHandlerBuilder mappingURL(String url){
    this._mappingUrl = url;
    return this;
  }

  Future<NUIErrorHandler> build() async {
    final db = await NUIDatabaseBuilder()
        .model(NUIErrorHandlerDatabase())
        .build();
    final errorHandler = NUIErrorHandler._builder(this);
    return errorHandler;
  }
}
